"""
Defines constants, specifications (classes with attributes defining varous parameters) and
containers for industry data generation and testing.

"""

from __future__ import annotations

import enum
from collections.abc import Sequence
from dataclasses import dataclass
from typing import ClassVar, Protocol

import numpy as np
from attrs import Attribute, Converter, cmp_using, field, validators
from attrs import frozen as frozen_attrs
from numpy.random import SeedSequence
from ruamel import yaml

from .. import (  # noqa: TID252
    DEFAULT_REC_RATIO,
    VERSION,
    ArrayBIGINT,
    ArrayBoolean,
    ArrayDouble,
    ArrayFloat,
    ArrayINT,
    RECForm,
    UPPAggrSelector,
    this_yaml,
)
from ..core.pseudorandom_numbers import (  # noqa: TID252
    DEFAULT_BETA_DIST_PARMS,
    DEFAULT_DIST_PARMS,
)

__version__ = VERSION


DEFAULT_FCOUNT_WTS = np.asarray((_nr := np.arange(6, 0, -1)) / _nr.sum(), float)


@dataclass(slots=True, frozen=True)
class SeedSequenceData:
    share: SeedSequence
    pcm: SeedSequence
    fcounts: SeedSequence | None
    price: SeedSequence | None


@this_yaml.register_class
@enum.unique
class PriceSpec(tuple[bool, str | None], enum.ReprEnum):
    """Price specification.

    Whether prices are symmetric and, if not, the direction of correlation, if any.
    """

    SYM = (True, None)
    ZERO = (False, None)
    NEG = (False, "negative share-correlation")
    POS = (False, "positive share-correlation")
    CSY = (False, "market-wide cost-symmetry")


@this_yaml.register_class
@enum.unique
class SHRDistribution(enum.StrEnum):
    """Market share distributions."""

    UNI = "Uniform"
    R"""Uniform distribution over :math:`s_1 + s_2 \leqslant 1`"""

    DIR_FLAT = "Flat Dirichlet"
    """Shape parameter for all merging-firm-shares is unity (1)"""

    DIR_FLAT_CONSTR = "Flat Dirichlet - Constrained"
    """Impose minimum probablility weight on each firm-count

    Only firm-counts with probability weight of 3% or more
    are included for data generation.
    """

    DIR_ASYM = "Asymmetric Dirichlet"
    """Share distribution for merging-firm shares has a higher peak share

    By default, shape parameter for merging-firm-share is 2.5, and
    1.0 for all others. Defining, :attr:`mergeron.ShareSpec.dist_parms`
    as a vector of shape parameters with length matching
    that of :attr:`mergeron.ShareSpec.dist_parms` allows flexible specification
    of Dirichlet-distributed share-data generation.
    """

    DIR_COND = "Conditional Dirichlet"
    """Shape parameters for non-merging firms is proportional

    Shape parameters for merging-firm-share are 2.0 each; and
    are equiproportional and add to 2.0 for all non-merging-firm-shares.
    """


def _fc_wts_conv(
    _v: Sequence[float | int] | ArrayDouble | ArrayINT | None, _i: ShareSpec
) -> ArrayFloat | None:
    if _i.dist_type == SHRDistribution.UNI:
        return None
    elif _v is None or np.array_equal(_v, DEFAULT_FCOUNT_WTS):
        return DEFAULT_FCOUNT_WTS
    else:
        return _tv if (_tv := np.asarray(_v, float)).sum() == 1 else _tv / _tv.sum()


def _shr_dp_conv(_v: Sequence[float] | ArrayFloat | None, _i: ShareSpec) -> ArrayFloat:
    match _v:
        case None if _i.dist_type == SHRDistribution.UNI:
            return DEFAULT_DIST_PARMS
        case None:
            _fc_max = 1 + (
                len(DEFAULT_FCOUNT_WTS)
                if not hasattr(_i, "firm_counts_weights")
                or _i.firm_counts_weights is None
                or len(_i.firm_counts_weights) == 0
                else len(_i.firm_counts_weights)
            )

            match _i.dist_type:
                case SHRDistribution.DIR_FLAT | SHRDistribution.DIR_FLAT_CONSTR:
                    return np.ones(_fc_max, float)
                case SHRDistribution.DIR_ASYM:
                    return np.array([2.0] * 6 + [1.5] * 5 + [1.25] * _fc_max, float)
                case SHRDistribution.DIR_COND:
                    return np.array([], float)
                case _ if isinstance(_i.dist_type, SHRDistribution):
                    raise ValueError(
                        f"No default defined for market share distribution, {_i.dist_type!r}"
                    )
                case _:
                    raise ValueError(
                        f"Unsupported distribution for market share generation, {_i.dist_type!r}"
                    )
        case _ if isinstance(_v, Sequence | np.ndarray):
            return np.asarray(_v, float)
        case _:
            raise ValueError(
                f"Input, {_v!r} has invalid type. Must be None, Sequence of floats, or Numpy ndarray."
            )


@this_yaml.register_class
@frozen_attrs
class ShareSpec:
    """Market share specification

    A salientfeature of market-share specification in this package is that
    the draws represent markets with multiple different firm-counts.
    Firm-counts are unspecified if the share distribution is
    :attr:`mergeron.SHRDistribution.UNI`, for Dirichlet-distributed market-shares,
    the default specification is that firm-counts  vary between
    2 and 7 firms with each value equally likely.

    Notes
    -----
    If :attr:`.dist_type` == :attr:`.SHRDistribution.UNI`, it is then infeasible that
    :attr:`.recapture_form` == :attr:`mergeron.RECForm.OUTIN`.
    In other words, recapture ratios cannot be estimated using
    outside-good choice probabilities if the distribution of markets over firm-counts
    is unspecified.

    """

    dist_type: SHRDistribution = field(kw_only=False)
    """See :class:`SHRDistribution`"""

    firm_counts_weights: ArrayFloat | None = field(
        kw_only=True,
        eq=cmp_using(eq=np.array_equal),
        converter=Converter(_fc_wts_conv, takes_self=True),  # type: ignore
    )
    """Relative or absolute frequencies of pre-merger firm counts

    Defaults to :attr:`DEFAULT_FCOUNT_WTS`, which specifies pre-merger
    firm-counts of 2 to 7 with weights in descending order from 6 to 1.

    ALERT: Firm-count weights are irrelevant when the mergering firmss shares are specified
    to have uniform distribution; therefore this attribute is forced to None if
    :attr:`.dist_type` == :attr:`.SHRDistribution.UNI`.

    """

    @firm_counts_weights.default
    def _fcwd(_i: ShareSpec) -> ArrayFloat | None:
        return _fc_wts_conv(None, _i)

    @firm_counts_weights.validator
    def _fcv(_i: ShareSpec, _a: Attribute[ArrayFloat], _v: ArrayFloat) -> None:
        if _i.dist_type != SHRDistribution.UNI and not len(_v):
            raise ValueError(
                f"Attribute, {'"firm_counts_weights"'} must be populated if the share distribution is "
                "other than uniform distribution."
            )

    dist_parms: ArrayFloat = field(
        kw_only=True,
        converter=Converter(_shr_dp_conv, takes_self=True),  # type: ignore
        eq=cmp_using(eq=np.array_equal),
    )
    """Parameters for tailoring market-share distribution

    For Uniform distribution, bounds of the distribution; defaults to `(0, 1)`;
    for Dirichlet-type distributions, a vector of shape parameters of length
    equal to 1 plus the length of firm-count weights below; defaults depend on
    type of Dirichlet-distribution specified.

    """

    @dist_parms.default
    def _dpd(_i: ShareSpec) -> ArrayFloat:
        # converters run after defaults, and we
        # avoid redundancy and confusion here
        return _shr_dp_conv(None, _i)

    @dist_parms.validator
    def _dpv(_i: ShareSpec, _a: Attribute[ArrayFloat], _v: ArrayFloat) -> None:
        if (
            _i.firm_counts_weights is not None
            and _v is not None
            and 0 < len(_v) < (1 + len(_i.firm_counts_weights))
        ):
            raise ValueError(
                "If specified, the number of distribution parameters must euqal or "
                "exceed the maximum firm-count premerger, which is "
                "1 plus the length of the vector specifying firm-count weights."
            )

    recapture_form: RECForm = field(default=RECForm.INOUT)
    """See :class:`mergeron.RECForm`"""

    @recapture_form.validator
    def _rfv(_i: ShareSpec, _a: Attribute[RECForm], _v: RECForm) -> None:
        if _i.dist_type == SHRDistribution.UNI and _v == RECForm.OUTIN:
            raise ValueError(
                "Outside-good choice probabilities cannot be generated if the "
                "merging firms' market shares have uniform distribution over the "
                "3-dimensional simplex with the distribution of markets over "
                "firm-counts unspecified."
            )

    recapture_ratio: float | None = field(kw_only=True)
    """A value between 0 and 1.

    :code:`None` if market share specification requires direct generation of
    outside good choice probabilities (:attr:`mergeron.RECForm.OUTIN`).

    The recapture ratio is usually calibrated to the numbers-equivalent of the
    HHI threshold for the presumtion of harm from unilateral competitive effects
    in published merger guidelines. Accordingly, the recapture ratio rounded to
    the nearest 5% is:

    * 0.85, **7-to-6 merger from symmetry**; US Guidelines, 1982, 1984, 1992, 2023
    * 0.80, 5-to-4 merger from symmetry
    * 0.80, **5-to-4 merger to symmetry**; US Guidelines, 2010

    Highlighting indicates hypothetical mergers in the neighborhood of (the boundary of)
    the Guidelines presumption of harm. (In the EU Guidelines, concentration measures serve as
    screens for further investigation, rather than as the basis for presumptions of harm or
    presumptions no harm.)

    ALERT: If diversion ratios are estimated by specifying a choice probability for the
    outside good, the recapture ratio is set to None, overriding any user-specified value.

    """

    @recapture_ratio.default
    def _rrd(_i: ShareSpec) -> float | None:
        return None if _i.recapture_form == RECForm.OUTIN else DEFAULT_REC_RATIO

    @recapture_ratio.validator
    def _rrv(_i: ShareSpec, _a: Attribute[float], _v: float) -> None:
        if _v and not (0 < _v <= 1):
            raise ValueError("Recapture ratio must lie in the interval, [0, 1).")
        elif _v is None and _i.recapture_form != RECForm.OUTIN:
            raise ValueError(
                f"Recapture specification, {_i.recapture_form!r} requires that "
                "the market sample specification inclues a recapture ratio in the "
                "interval [0, 1)."
            )

    @classmethod
    def to_yaml(
        cls, _r: yaml.representer.SafeRepresenter, _d: ShareSpec
    ) -> yaml.MappingNode:
        _ret: yaml.MappingNode = _r.represent_mapping(
            f"!{cls.__name__}",
            {_a.name: getattr(_d, _a.name) for _a in _d.__attrs_attrs__},
        )
        return _ret

    @classmethod
    def from_yaml(
        cls, _c: yaml.constructor.SafeConstructor, _n: yaml.MappingNode
    ) -> ShareSpec:
        return cls(**_c.construct_mapping(_n))


@enum.unique
class PCMDistribution(enum.StrEnum):
    """Margin distributions."""

    UNI = "Uniform"
    BETA = "Beta"
    BETA_BND = "Bounded Beta"
    EMPR = "Damodaran margin data, resampled"


@enum.unique
class FM2Constraint(enum.StrEnum):
    """Firm 2 margins - derivation methods."""

    IID = "i.i.d"
    MNL = "MNL-dep"
    SYM = "symmetric"


@this_yaml.register_class
@frozen_attrs
class PCMSpec:
    """Price-cost margin (PCM) specification

    If price-cost margins are specified as having Beta distribution,
    `dist_parms` is specified as a pair of positive, non-zero shape parameters of
    the standard Beta distribution. Specifying shape parameters :code:`np.array([1, 1])`
    is known equivalent to specifying uniform distribution over
    the interval :math:`[0, 1]`. If price-cost margins are specified as having
    Bounded-Beta distribution, `dist_parms` is specified as
    the tuple, (`mean`, `std deviation`, `min`, `max`), where `min` and `max`
    are lower- and upper-bounds respectively within the interval :math:`[0, 1]`.


    """

    dist_type: PCMDistribution = field()
    """See :class:`PCMDistribution`"""

    @dist_type.default
    def _dtd(_i: PCMSpec) -> PCMDistribution:
        return PCMDistribution.UNI

    dist_parms: ArrayFloat | None = field(kw_only=True, eq=np.array_repr)
    """Parameter specification for tailoring PCM distribution

    For Uniform distribution, bounds of the distribution; defaults to `(0, 1)`;
    for Beta distribution, shape parameters, defaults to `(1, 1)`;
    for Bounded-Beta distribution, vector of (min, max, mean, std. deviation), non-optional;
    for empirical distribution based on Damodaran margin data, optional, ignored

    """

    @dist_parms.default
    def _dpwd(_i: PCMSpec) -> ArrayFloat | None:
        match _i.dist_type:
            case PCMDistribution.EMPR:
                return None
            case PCMDistribution.BETA:
                return DEFAULT_BETA_DIST_PARMS
            case PCMDistribution.BETA_BND:
                return np.array([0.5, 1.0, 0.0, 0.1], float)
            case _:
                return DEFAULT_DIST_PARMS

    @dist_parms.validator
    def _dpv(
        _i: PCMSpec, _a: Attribute[ArrayFloat | None], _v: ArrayFloat | None
    ) -> None:
        if _i.dist_type.name.startswith("BETA"):
            if _v is None or not any(_v.shape):
                pass
            elif np.array_equal(_v, DEFAULT_DIST_PARMS):
                raise ValueError(
                    f"The distribution parameters, {DEFAULT_DIST_PARMS!r} "
                    "are not valid with margin distribution, {_dist_type_pcm!r}"
                )
            elif (
                _i.dist_type == PCMDistribution.BETA and len(_v) != len(("a", "b"))
            ) or (
                _i.dist_type == PCMDistribution.BETA_BND
                and len(_v) != len(("mu", "sigma", "max", "min"))
            ):
                raise ValueError(
                    f"Given number, {len(_v)} of parameters "
                    f'for PCM with distribution, "{_i.dist_type}" is incorrect.'
                )

        elif _v is not None and _i.dist_type == PCMDistribution.EMPR:
            raise ValueError(
                f"Empirical distribution does not require additional parameters; "
                f'"given value, {_v!r} is ignored."'
            )

    firm2_pcm_constraint: FM2Constraint = field(kw_only=True, default=FM2Constraint.IID)
    """See :class:`FM2Constraint`"""

    @classmethod
    def to_yaml(
        cls, _r: yaml.representer.SafeRepresenter, _d: PCMSpec
    ) -> yaml.MappingNode:
        _ret: yaml.MappingNode = _r.represent_mapping(
            f"!{cls.__name__}",
            {_a.name: getattr(_d, _a.name) for _a in _d.__attrs_attrs__},
        )
        return _ret

    @classmethod
    def from_yaml(
        cls, _c: yaml.constructor.SafeConstructor, _n: yaml.MappingNode
    ) -> PCMSpec:
        return cls(**_c.construct_mapping(_n))


@enum.unique
class SSZConstant(float, enum.ReprEnum):
    """
    Scale factors to offset sample size reduction.

    Sample size reduction occurs when imposing a HSR filing test
    or equilibrium condition under MNL demand.
    """

    HSR_NTH = 1.666667
    """
    For HSR filing requirement.

    When filing requirement is assumed met if maximum merging-firm shares exceeds
    ten (10) times the n-th firm's share and minimum merging-firm share is
    no less than n-th firm's share. To assure that the number of draws available
    after applying the given restriction, the initial number of draws is larger than
    the sample size by the given scale factor.
    """

    HSR_TEN = 1.234567
    """
    For alternative HSR filing requirement,

    When filing requirement is assumed met if merging-firm shares exceed 10:1 ratio
    to each other.
    """

    MNL_DEP = 1.25
    """
    For restricted PCM's.

    When merging firm's PCMs are constrained for consistency with f.o.c.s from
    profit maximization under Nash-Bertrand oligopoly with MNL demand.
    """

    ONE = 1.00
    """When initial set of draws is not restricted in any way."""


@dataclass(slots=True, frozen=True)
class MarketSampleData:
    """Container for generated markets data sample."""

    frmshr_array: ArrayDouble
    """Merging-firm shares (with two merging firms)"""

    pcm_array: ArrayDouble
    """Merging-firms' prices (normalized to 1, in default specification)"""

    price_array: ArrayDouble
    """Merging-firms' price-cost margins (PCM)"""

    fcounts: ArrayBIGINT
    """Number of firms in market"""

    aggregate_purchase_prob: ArrayDouble
    """
    One (1) minus probability that the outside good is chosen

    Converts market shares to choice probabilities by multiplication.
    """

    nth_firm_share: ArrayDouble
    """Market-share of n-th firm

    Relevant for testing for draws the do or
    do not meet HSR filing thresholds.
    """

    divr_array: ArrayDouble
    """Diversion ratio between the merging firms"""

    hhi_post: ArrayDouble
    """Post-merger change in Herfindahl-Hirschmann Index (HHI)"""

    hhi_delta: ArrayDouble
    """Change in HHI from combination of merging firms"""


@dataclass(slots=True, frozen=True)
class ShareDataSample:
    """Container for generated market shares.

    Includes related measures of market structure
    and aggregate purchase probability.
    """

    mktshr_array: ArrayDouble
    """All-firm shares (with two merging firms)"""

    fcounts: ArrayBIGINT
    """All-firm-count for each draw"""

    nth_firm_share: ArrayDouble
    """Market-share of n-th firm"""

    aggregate_purchase_prob: ArrayDouble
    """Converts market shares to choice probabilities by multiplication."""


@dataclass(slots=True, frozen=True)
class PriceDataSample:
    """Container for generated price array, and related."""

    price_array: ArrayDouble
    """Merging-firms' prices"""

    hsr_filing_test: ArrayBoolean
    """Flags draws as meeting HSR filing thresholds or not"""


@dataclass(slots=True, frozen=True)
class MarginDataSample:
    """Container for generated margin array and related MNL test array."""

    pcm_array: ArrayDouble
    """Merging-firms' PCMs"""

    mnl_test_array: ArrayBoolean
    """Flags infeasible observations as False and rest as True

    Applying restrictions from Bertrand-Nash oligopoly
    with MNL demand results in draws of Firm 2 PCM falling
    outside the feabile interval,:math:`[0, 1]`, depending
    on the configuration of merging firm shares. Such draws
    are are flagged as infeasible (False)in :code:`mnl_test_array` while
    draws with PCM values within the feasible range are
    flagged as True. Used from filtering-out draws with
    infeasible PCM.
    """


@enum.unique
class INVResolution(enum.StrEnum):
    CLRN = "clearance"
    ENFT = "enforcement"
    BOTH = "investigation"


@frozen_attrs
class UPPTestRegime:
    """Configuration for UPP tests."""

    resolution: INVResolution = field(
        kw_only=False,
        default=INVResolution.ENFT,
        validator=validators.in_([INVResolution.CLRN, INVResolution.ENFT]),
    )
    """Whether to test clearance, enforcement, or both."""

    guppi_aggregator: UPPAggrSelector = field(
        kw_only=False, default=UPPAggrSelector.MIN
    )
    """Aggregator for GUPPI test."""

    divr_aggregator: UPPAggrSelector = field(kw_only=False, default=UPPAggrSelector.MIN)
    """Aggregator for diversion ratio test."""


@dataclass(slots=True, frozen=True)
class UPPTestsRaw:
    """Container for arrays marking test failures and successes

    A test success is a draw ("market") that meeets the
    specified test criterion, and a test failure is
    one that does not; test criteria are evaluated in
    :func:`enforcement_stats.gen_upp_arrays`.
    """

    guppi_test_simple: ArrayBoolean
    """True if GUPPI estimate meets criterion"""

    guppi_test_compound: ArrayBoolean
    """True if both GUPPI estimate and diversion ratio estimate
    meet criterion
    """

    cmcr_test: ArrayBoolean
    """True if CMCR estimate meets criterion"""

    ipr_test: ArrayBoolean
    """True if IPR (partial price-simulation) estimate meets criterion"""


@dataclass(slots=True, frozen=True)
class UPPTestsCounts:
    """Counts of markets resolved as specified

    Resolution may be either :attr:`INVResolution.ENFT`,
    :attr:`INVResolution.CLRN`, or :attr:`INVResolution.BOTH`.
    In the case of :attr:`INVResolution.BOTH`, two colums of counts
    are returned: one for each resolution.

    """

    by_firm_count: ArrayBIGINT
    by_delta: ArrayBIGINT
    by_conczone: ArrayBIGINT
    """Zones are "unoncentrated", "moderately concentrated", and "highly concentrated",
    with futher detail by HHI and ΔHHI for mergers in the "unconcentrated" and
    "moderately concentrated" zones. See
    :attr:`mergeron.gen.enforcement_stats.HMG_PRESUMPTION_ZONE_MAP` and
    :attr:`mergeron.gen.enforcement_stats.ZONE_VALS` for more detail.

    """


# https://stackoverflow.com/questions/54668000
class DataclassInstance(Protocol):
    """Generic dataclass-instance"""

    __dataclass_fields__: ClassVar


for _typ in (
    PriceSpec,
    SHRDistribution,
    PCMDistribution,
    FM2Constraint,
    SSZConstant,
    INVResolution,
):
    #  NOTE: If additional enums are defined in this module,
    #  add themn to the list above

    _, _ = (
        this_yaml.representer.add_representer(
            _typ,
            lambda _r, _d: _r.represent_scalar(f"!{_d.__class__.__name__}", _d.name),
        ),
        this_yaml.constructor.add_constructor(
            f"!{_typ.__name__}",
            lambda _c, _n, /: getattr(
                globals().get(_n.tag.lstrip("!")), _c.construct_scalar(_n)
            ),
        ),
    )

for _typ in (MarketSampleData, SeedSequenceData, UPPTestsCounts):
    _, _ = (
        this_yaml.representer.add_representer(
            _typ,
            lambda _r, _d: _r.rpresent_mapping(
                f"!{_d.__class__.__name__}",
                {_a: getattr(_d, _a) for _a in _d.__dataclass_fields__},
            ),
        ),
        this_yaml.constructor.add_constructor(
            _typ,
            lambda _c, _n: globals().get(
                _n.tag.lstrip("!")(**_c.construct_mapping(_n))
            ),
        ),
    )
